from dateutil.rrule import FR
from django.test import TestCase

from schedule.models import Rule


class TestPeriod(TestCase):
    def test_get_params(self):
        rule = Rule(
            params="count:1;bysecond:1;byminute:1,2,4,5;byweekday:FR;bysetpos:-1"
        )
        expected = {
            "count": 1,
            "byminute": [1, 2, 4, 5],
            "bysecond": 1,
            "byweekday": FR,
            "bysetpos": -1,
        }
        self.assertEqual(rule.get_params(), expected)



import datetime
import json

import pytz
from django.contrib.auth.models import User
from django.http import Http404
from django.test import RequestFactory, SimpleTestCase, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone

from schedule.models.calendars import Calendar
from schedule.models.events import Event, Occurrence
from schedule.models.rules import Rule
from schedule.settings import USE_FULLCALENDAR
from schedule.views import (
    check_next_url,
    coerce_date_dict,
    get_next_url,
    get_occurrence,
)


class TestViews(TestCase):
    fixtures = ["schedule.json"]

    def setUp(self):
        self.rule = Rule.objects.create(frequency="DAILY")
        self.calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        self.event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=self.rule,
            calendar=self.calendar,
        )

    @override_settings(USE_TZ=False)
    def test_timezone_off(self):
        url = reverse("day_calendar", kwargs={"calendar_slug": self.calendar.slug})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)


class TestViewUtils(TestCase):
    def setUp(self):
        self.rule = Rule.objects.create(frequency="DAILY")
        self.calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        self.event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=self.rule,
            calendar=self.calendar,
        )

    def test_get_occurrence(self):
        event, occurrence = get_occurrence(
            self.event.pk,
            year=2008,
            month=1,
            day=5,
            hour=8,
            minute=0,
            second=0,
            tzinfo=pytz.utc,
        )
        self.assertEqual(event, self.event)
        self.assertEqual(occurrence.start, self.event.start)
        self.assertEqual(occurrence.end, self.event.end)

    def test_get_occurrence_raises(self):
        with self.assertRaises(Http404):
            get_occurrence(
                self.event.pk,
                year=2007,
                month=1,
                day=5,
                hour=8,
                minute=0,
                second=0,
                tzinfo=pytz.utc,
            )

    def test_get_occurrence_persisted(self):
        date = timezone.make_aware(
            datetime.datetime(year=2008, month=1, day=5, hour=8, minute=0, second=0),
            pytz.utc,
        )
        occurrence = self.event.get_occurrence(date)
        occurrence.save()
        with self.assertRaises(Http404):
            get_occurrence(self.event.pk, occurrence_id=100)

        event, persisted_occ = get_occurrence(
            self.event.pk, occurrence_id=occurrence.pk
        )
        self.assertEqual(persisted_occ, occurrence)

    @override_settings(TIME_ZONE="America/Montevideo")
    def test_get_occurrence_raises_wrong_tz(self):
        # Montevideo is 3 hours behind UTC
        with self.assertRaises(Http404):
            event, occurrence = get_occurrence(
                self.event.pk, year=2008, month=1, day=5, hour=8, minute=0, second=0
            )

    def test_coerce_date_dict(self):
        self.assertEqual(
            coerce_date_dict(
                {
                    "year": "2008",
                    "month": "4",
                    "day": "2",
                    "hour": "4",
                    "minute": "4",
                    "second": "4",
                }
            ),
            {"year": 2008, "month": 4, "day": 2, "hour": 4, "minute": 4, "second": 4},
        )

    def test_coerce_date_dict_partial(self):
        self.assertEqual(
            coerce_date_dict({"year": "2008", "month": "4", "day": "2"}),
            {"year": 2008, "month": 4, "day": 2, "hour": 0, "minute": 0, "second": 0},
        )

    def test_coerce_date_dict_empty(self):
        self.assertEqual(coerce_date_dict({}), {})

    def test_coerce_date_dict_missing_values(self):
        self.assertEqual(
            coerce_date_dict({"year": "2008", "month": "4", "hours": "3"}),
            {"year": 2008, "month": 4, "day": 1, "hour": 0, "minute": 0, "second": 0},
        )


class TestGetNextUrl(SimpleTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

    def test_redirects_to_same_server(self):
        redirect_to = "http://testserver/"
        request = self.factory.get(f"?next={redirect_to}")
        self.assertEqual(get_next_url(request, None), redirect_to)

    def test_redirects_to_malicious_server(self):
        request = self.factory.get("?next=http://evil.com")
        self.assertIsNone(get_next_url(request, None))


class TestUrls(TestCase):
    fixtures = ["schedule.json"]
    highest_event_id = 7

    def test_calendar_view(self):
        response = self.client.get(
            reverse("year_calendar", kwargs={"calendar_slug": "example"}), {}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context[0]["calendar"].name, "Example Calendar")

    def test_calendar_month_view(self):
        response = self.client.get(
            reverse("month_calendar", kwargs={"calendar_slug": "example"}),
            {"year": 2000, "month": 11},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context[0]["calendar"].name, "Example Calendar")
        month = response.context[0]["period"]
        self.assertEqual(
            (month.start, month.end),
            (
                datetime.datetime(2000, 11, 1, 0, 0, tzinfo=pytz.utc),
                datetime.datetime(2000, 12, 1, 0, 0, tzinfo=pytz.utc),
            ),
        )

    def test_event_creation_anonymous_user(self):
        response = self.client.get(
            reverse("calendar_create_event", kwargs={"calendar_slug": "example"})
        )
        self.assertEqual(response.status_code, 302)

    def test_event_creation_authenticated_user(self):
        user = User.objects.create_user("event_create_admin", password="admin")
        self.client.force_login(user)
        response = self.client.get(
            reverse("calendar_create_event", kwargs={"calendar_slug": "example"})
        )

        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("calendar_create_event", kwargs={"calendar_slug": "example"}),
            {
                "description": "description",
                "title": "title",
                "end_recurring_period_1": "10:22:00",
                "end_recurring_period_0": "2008-10-30",
                "end_recurring_period_2": "AM",
                "end_1": "10:22:00",
                "end_0": "2008-10-30",
                "end_2": "AM",
                "start_0": "2008-10-30",
                "start_1": "09:21:57",
                "start_2": "AM",
            },
        )
        self.assertEqual(response.status_code, 302)

        highest_event_id = self.highest_event_id
        highest_event_id += 1
        response = self.client.get(
            reverse("event", kwargs={"event_id": highest_event_id})
        )
        self.assertEqual(response.status_code, 200)

    def test_view_event(self):
        response = self.client.get(reverse("event", kwargs={"event_id": 1}))
        self.assertEqual(response.status_code, 200)

    def test_delete_event_anonymous_user(self):
        # Only logged-in users should be able to delete, so we're redirected
        response = self.client.get(reverse("delete_event", kwargs={"event_id": 1}))
        self.assertEqual(response.status_code, 302)

    def test_delete_event_authenticated_user(self):
        user = User.objects.create_user("event_delete_admin", password="admin")
        self.client.force_login(user)
        # Load the deletion page
        response = self.client.get(reverse("delete_event", kwargs={"event_id": 1}))
        self.assertEqual(response.status_code, 200)
        if USE_FULLCALENDAR:
            self.assertEqual(
                response.context["next"],
                reverse("fullcalendar", args=[Event.objects.get(id=1).calendar.slug]),
            )
        else:
            self.assertEqual(
                response.context["next"],
                reverse("day_calendar", args=[Event.objects.get(id=1).calendar.slug]),
            )

        # Delete the event
        response = self.client.post(reverse("delete_event", kwargs={"event_id": 1}))
        self.assertEqual(response.status_code, 302)

        # Since the event is now deleted, we get a 404
        response = self.client.get(reverse("delete_event", kwargs={"event_id": 1}))
        self.assertEqual(response.status_code, 404)

    def test_occurrences_api_returns_the_expected_occurrences(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        rule = Rule.objects.create(frequency="DAILY")
        Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=calendar,
        )
        # test calendar slug
        response = self.client.get(
            reverse("api_occurrences")
            + "?calendar={}&start={}&end={}".format(
                "MyCal", datetime.datetime(2008, 1, 5), datetime.datetime(2008, 1, 6)
            )
        )
        self.assertEqual(response.status_code, 200)
        expected_content = [
            {
                "existed": False,
                "end": "2008-01-05T09:00:00Z",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-05-05T00:00:00Z",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-05T08:00:00Z",
                "id": 9,
            }
        ]
        self.assertEqual(json.loads(response.content.decode()), expected_content)

    def test_occurrences_api_without_parameters_return_status_400(self):
        response = self.client.get(reverse("api_occurrences"))
        self.assertEqual(response.status_code, 400)

    def test_occurrences_api_without_calendar_slug_return_status_404(self):
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": datetime.datetime(2008, 1, 5),
                "end": datetime.datetime(2008, 1, 6),
                "calendar_slug": "NoMatch",
            },
        )
        self.assertEqual(response.status_code, 400)

    def test_occurrences_api_checks_valid_occurrence_ids(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        rule = Rule.objects.create(frequency="DAILY")
        event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 1, 8, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=calendar,
        )
        Occurrence.objects.create(
            event=event,
            title="My persisted Occ",
            description="Persisted occ test",
            start=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
            original_start=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
            original_end=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
        )
        # test calendar slug
        response = self.client.get(
            reverse("api_occurrences")
            + "?calendar={}&start={}&end={}".format(
                "MyCal", datetime.datetime(2008, 1, 5), datetime.datetime(2008, 1, 8)
            )
        )
        self.assertEqual(response.status_code, 200)
        expected_content = [
            {
                "existed": False,
                "end": "2008-01-05T09:00:00Z",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-08T00:00:00Z",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-05T08:00:00Z",
                "id": 10,
            },
            {
                "existed": False,
                "end": "2008-01-06T09:00:00Z",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-08T00:00:00Z",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-06T08:00:00Z",
                "id": 10,
            },
            {
                "existed": False,
                "end": "2008-01-07T09:00:00Z",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-08T00:00:00Z",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-07T08:00:00Z",
                "id": 10,
            },
            {
                "existed": True,
                "end": "2008-01-07T08:00:00Z",
                "description": "Persisted occ test",
                "creator": "None",
                "color": "",
                "title": "My persisted Occ",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-08T00:00:00Z",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-07T08:00:00Z",
                "id": 1,
            },
        ]
        self.assertEqual(json.loads(response.content.decode()), expected_content)
        # test timezone param
        response = self.client.get(
            reverse("api_occurrences")
            + "?calendar={}&start={}&end={}&timezone={}".format(
                "MyCal",
                datetime.datetime(2008, 1, 5),
                datetime.datetime(2008, 1, 8),
                "America/Chicago",
            )
        )
        self.assertEqual(response.status_code, 200)
        expected_content = [
            {
                "existed": False,
                "end": "2008-01-05T03:00:00-06:00",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-07T18:00:00-06:00",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-05T02:00:00-06:00",
                "id": 10,
            },
            {
                "existed": False,
                "end": "2008-01-06T03:00:00-06:00",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-07T18:00:00-06:00",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-06T02:00:00-06:00",
                "id": 10,
            },
            {
                "existed": False,
                "end": "2008-01-07T03:00:00-06:00",
                "description": "",
                "creator": "None",
                "color": "",
                "title": "Recent Event",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-07T18:00:00-06:00",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-07T02:00:00-06:00",
                "id": 10,
            },
            {
                "existed": True,
                "end": "2008-01-07T02:00:00-06:00",
                "description": "Persisted occ test",
                "creator": "None",
                "color": "",
                "title": "My persisted Occ",
                "rule": "",
                "event_id": 8,
                "end_recurring_period": "2008-01-07T18:00:00-06:00",
                "cancelled": False,
                "calendar": "MyCalSlug",
                "start": "2008-01-07T02:00:00-06:00",
                "id": 1,
            },
        ]
        self.assertEqual(json.loads(response.content.decode()), expected_content)

    def test_occurrences_api_works_with_and_without_cal_slug(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            calendar=calendar,
        )
        # test calendar slug
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2008-01-05",
                "end": "2008-02-05",
                "calendar_slug": event.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event.title, [d["title"] for d in resp_list])
        # test works with no calendar slug
        response = self.client.get(
            reverse("api_occurrences"), {"start": "2008-01-05", "end": "2008-02-05"}
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event.title, [d["title"] for d in resp_list])

    def test_cal_slug_filters_returned_events(self):
        calendar1 = Calendar.objects.create(name="MyCal1", slug="MyCalSlug1")
        calendar2 = Calendar.objects.create(name="MyCal2", slug="MyCalSlug2")
        event1 = Event.objects.create(
            title="Recent Event 1",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            calendar=calendar1,
        )
        event2 = Event.objects.create(
            title="Recent Event 2",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            calendar=calendar2,
        )
        # Test both present with no cal arg
        response = self.client.get(
            reverse("api_occurrences"), {"start": "2008-01-05", "end": "2008-02-05"}
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event1.title, [d["title"] for d in resp_list])
        self.assertIn(event2.title, [d["title"] for d in resp_list])
        # test event2 not in event1 response
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2008-01-05",
                "end": "2008-02-05",
                "calendar_slug": event1.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event1.title, [d["title"] for d in resp_list])
        self.assertNotIn(event2.title, [d["title"] for d in resp_list])

    def test_occurrences_api_works_with_different_date_string_formats(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            calendar=calendar,
        )
        # test works with date string time format '%Y-%m-%d'
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2008-01-05",
                "end": "2008-02-05",
                "calendar_slug": event.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event.title, [d["title"] for d in resp_list])
        # test works with date string time format '%Y-%m-%dT%H:%M:%S'
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2008-01-05T00:00:00",
                "end": "2008-02-05T00:00:00",
                "calendar_slug": event.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 200)
        resp_list = json.loads(response.content.decode())
        self.assertIn(event.title, [d["title"] for d in resp_list])

    def test_occurrences_api_fails_with_incorrect_date_string_formats(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            calendar=calendar,
        )

        # test fails with date string time format not '%Y-%m-%d' or '%Y-%m-%dT%H:%M:%S'
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2008-01-05T00:00",
                "end": "2008-02-05T00:00",
                "calendar_slug": event.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 400)
        resp = response.content.decode()
        expected_error = "does not match format '%Y-%m-%dT%H:%M:%S'"
        self.assertIn(expected_error, resp)

    def test_occurrences_api_filters_cancelled_events(self):
        # create a calendar and event
        calendar = Calendar.objects.create(name="MyCal", slug="MyCalSlug")
        weekly_meeting_event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2021, 12, 27, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2021, 12, 27, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2021, 12, 31, 0, 0, tzinfo=pytz.utc),
            calendar=calendar,
        )
        Occurrence.objects.create(
            event=weekly_meeting_event,
            start=weekly_meeting_event.start.replace(year=2021, month=12, day=27),
            end=weekly_meeting_event.end.replace(year=2021, month=12, day=27),
            cancelled=True,
            original_start=weekly_meeting_event.start.replace(
                year=2021, month=12, day=27
            ),
            original_end=weekly_meeting_event.end.replace(year=2021, month=12, day=27),
        )

        # test fails with date string time format not '%Y-%m-%d' or '%Y-%m-%dT%H:%M:%S'
        response = self.client.get(
            reverse("api_occurrences"),
            {
                "start": "2021-12-27T08:00:00",
                "end": "2021-12-27T09:00:00",
                "calendar_slug": weekly_meeting_event.calendar.slug,
            },
        )
        self.assertEqual(response.status_code, 200)
        expected_content = []
        self.assertEqual(json.loads(response.content.decode()), expected_content)

    def test_check_next_url_valid_case(self):
        expected = "/calendar/1"
        res = check_next_url("/calendar/1")
        self.assertEqual(expected, res)

    def test_check_next_url_invalid_case(self):
        expected = None
        res = check_next_url("http://localhost/calendar/1")
        self.assertEqual(expected, res)
        res = check_next_url(None)
        self.assertEqual(expected, res)

    @override_settings(SITE_ID=1)
    def test_feed_link(self):
        feed_url = reverse("upcoming_events_feed", kwargs={"calendar_id": 1})
        response = self.client.get(feed_url)
        self.assertEqual(response.status_code, 200)
        expected_feed = "http://example.com/feed/calendar/upcoming/1/"
        self.assertTrue(expected_feed in response.content.decode())

    def test_calendar_view_home(self):
        calendar_view_url = reverse(
            "calendar_home", kwargs={"calendar_slug": "example"}
        )
        response = self.client.get(calendar_view_url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            '<a href="/feed/calendar/upcoming/1/">Feed</a>' in response.content.decode()
        )



import datetime

from django.test import TestCase
from django.utils import timezone

from schedule.models import Calendar, Event, Occurrence, Rule
from schedule.utils import EventListManager, OccurrenceReplacer


class TestEventListManager(TestCase):
    def setUp(self):
        weekly = Rule.objects.create(frequency="WEEKLY")
        daily = Rule.objects.create(frequency="DAILY")
        cal = Calendar.objects.create(name="MyCal")
        self.default_tzinfo = timezone.get_default_timezone()

        self.event1 = Event.objects.create(
            title="Weekly Event",
            start=datetime.datetime(2009, 4, 1, 8, 0, tzinfo=self.default_tzinfo),
            end=datetime.datetime(2009, 4, 1, 9, 0, tzinfo=self.default_tzinfo),
            end_recurring_period=datetime.datetime(
                2009, 10, 5, 0, 0, tzinfo=self.default_tzinfo
            ),
            rule=weekly,
            calendar=cal,
        )
        self.event2 = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=self.default_tzinfo),
            end=datetime.datetime(2008, 1, 5, 10, 0, tzinfo=self.default_tzinfo),
            end_recurring_period=datetime.datetime(
                2009, 5, 5, 0, 0, tzinfo=self.default_tzinfo
            ),
            rule=daily,
            calendar=cal,
        )

    def test_occurrences_after(self):
        eml = EventListManager([self.event1, self.event2])
        occurrences = eml.occurrences_after(
            datetime.datetime(2009, 4, 1, 0, 0, tzinfo=self.default_tzinfo)
        )
        self.assertEqual(next(occurrences).event, self.event1)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event2)
        self.assertEqual(next(occurrences).event, self.event1)
        occurrences = eml.occurrences_after()
        self.assertEqual(list(occurrences), [])


class TestOccurrenceReplacer(TestCase):
    def setUp(self):
        weekly = Rule.objects.create(frequency="WEEKLY")
        daily = Rule.objects.create(frequency="DAILY")
        cal = Calendar.objects.create(name="MyCal")
        self.default_tzinfo = timezone.get_default_timezone()
        self.start = timezone.now() - datetime.timedelta(days=10)
        self.end = self.start + datetime.timedelta(days=300)
        self.event1 = Event.objects.create(
            title="Weekly Event",
            start=self.start,
            end=self.end,
            end_recurring_period=self.end,
            rule=weekly,
            calendar=cal,
        )
        self.occ = Occurrence.objects.create(
            event=self.event1,
            start=self.start,
            end=self.end,
            original_start=self.start,
            original_end=self.end,
        )

        self.event2 = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=self.default_tzinfo),
            end=datetime.datetime(2008, 1, 5, 10, 0, tzinfo=self.default_tzinfo),
            end_recurring_period=datetime.datetime(
                2009, 5, 5, 0, 0, tzinfo=self.default_tzinfo
            ),
            rule=daily,
            calendar=cal,
        )

    def test_has_occurrence(self):
        other_occ = Occurrence.objects.create(
            event=self.event1,
            start=self.start,
            end=self.end,
            original_start=self.start,
            original_end=self.end,
        )
        occ_replacer = OccurrenceReplacer([self.occ])

        self.assertTrue(occ_replacer.has_occurrence(self.occ))
        self.assertTrue(occ_replacer.has_occurrence(other_occ))

    def test_has_occurrence_with_other_event(self):
        other_occ = Occurrence.objects.create(
            event=self.event2,
            start=self.start,
            end=self.end,
            original_start=self.start,
            original_end=self.end,
        )
        occ_replacer = OccurrenceReplacer([self.occ])

        self.assertTrue(occ_replacer.has_occurrence(self.occ))
        self.assertFalse(occ_replacer.has_occurrence(other_occ))

    def test_get_additional_occurrences(self):
        occ_replacer = OccurrenceReplacer([self.occ])
        # Other occurrence.
        Occurrence.objects.create(
            event=self.event2,
            start=self.start + datetime.timedelta(days=5),
            end=self.end,
            original_start=self.start,
            original_end=self.end,
        )
        res = occ_replacer.get_additional_occurrences(self.start, self.end)
        self.assertEqual(res, [self.occ])

    def test_get_additional_occurrences_cancelled(self):
        occ_replacer = OccurrenceReplacer([self.occ])
        self.occ.cancelled = True
        self.occ.save()
        res = occ_replacer.get_additional_occurrences(self.start, self.end)
        self.assertEqual(res, [])

    def test_get_occurrence(self):
        # self.occ is a persisted Occurrence
        occ_replacer = OccurrenceReplacer([self.occ])
        res = occ_replacer.get_occurrence(self.occ)
        self.assertEqual(res, self.occ)
        res = occ_replacer.get_occurrence(self.occ)
        self.assertEqual(res, self.occ)

    def test_get_occurrence_works_for_event_like_object(self):
        # get_occurrence method checks the duck
        occ_replacer = OccurrenceReplacer([self.occ])
        with self.assertRaises(AttributeError):
            occ_replacer.get_occurrence(int)



import datetime
from urllib.parse import urlencode

from django import template
from django.conf import settings
from django.template.loader import get_template
from django.urls import reverse
from django.utils import timezone
from django.utils.dateformat import format
from django.utils.html import escape

from schedule.models import Calendar
from schedule.periods import weekday_abbrs, weekday_names
from schedule.settings import (
    CHECK_CALENDAR_PERM_FUNC,
    CHECK_EVENT_PERM_FUNC,
    SCHEDULER_PREVNEXT_LIMIT_SECONDS,
)

register = template.Library()


@register.inclusion_tag("schedule/_month_table.html", takes_context=True)
def month_table(context, calendar, month, size="regular", shift=None):
    if shift:
        if shift == -1:
            month = month.prev()
        if shift == 1:
            month = next(month)
    if size == "small":
        context["day_names"] = weekday_abbrs
    else:
        context["day_names"] = weekday_names
    context["calendar"] = calendar
    context["month"] = month
    context["size"] = size
    return context


@register.inclusion_tag("schedule/_day_cell.html", takes_context=True)
def day_cell(context, calendar, day, month, size="regular"):
    context.update({"calendar": calendar, "day": day, "month": month, "size": size})
    return context


@register.inclusion_tag("schedule/_daily_table.html", takes_context=True)
def daily_table(context, day, start=8, end=20, increment=30):
    """
    Display a nice table with occurrences and action buttons.
    Arguments:
    start - hour at which the day starts
    end - hour at which the day ends
    increment - size of a time slot (in minutes)
    """
    user = context["request"].user
    addable = CHECK_EVENT_PERM_FUNC(None, user)
    if "calendar" in context:
        addable = addable and CHECK_CALENDAR_PERM_FUNC(context["calendar"], user)
    context["addable"] = addable
    day_part = day.get_time_slot(
        day.start + datetime.timedelta(hours=start),
        day.start + datetime.timedelta(hours=end),
    )
    # get slots to display on the left
    slots = _cook_slots(day_part, increment)
    context["slots"] = slots
    return context


@register.inclusion_tag("schedule/_event_title.html", takes_context=True)
def title(context, occurrence):
    context.update({"occurrence": occurrence})
    return context


@register.inclusion_tag("schedule/_event_options.html", takes_context=True)
def options(context, occurrence):
    context.update(
        {"occurrence": occurrence, "MEDIA_URL": getattr(settings, "MEDIA_URL")}
    )
    context["view_occurrence"] = occurrence.get_absolute_url()
    user = context["request"].user
    if CHECK_EVENT_PERM_FUNC(occurrence.event, user) and CHECK_CALENDAR_PERM_FUNC(
        occurrence.event.calendar, user
    ):
        context["edit_occurrence"] = occurrence.get_edit_url()
        context["cancel_occurrence"] = occurrence.get_cancel_url()
        context["delete_event"] = reverse("delete_event", args=(occurrence.event.id,))
        context["edit_event"] = reverse(
            "edit_event", args=(occurrence.event.calendar.slug, occurrence.event.id)
        )
    else:
        context["edit_event"] = context["delete_event"] = ""
    return context


@register.inclusion_tag("schedule/_create_event_options.html", takes_context=True)
def create_event_url(context, calendar, slot):
    context.update({"calendar": calendar, "MEDIA_URL": getattr(settings, "MEDIA_URL")})
    lookup_context = {"calendar_slug": calendar.slug}
    context["create_event_url"] = "{}{}".format(
        reverse("calendar_create_event", kwargs=lookup_context),
        querystring_for_date(slot),
    )
    return context


class CalendarNode(template.Node):
    def __init__(self, content_object, distinction, context_var, create=False):
        self.content_object = template.Variable(content_object)
        self.distinction = distinction
        self.context_var = context_var

    def render(self, context):
        Calendar.objects.get_calendar_for_object(
            self.content_object.resolve(context), self.distinction
        )
        context[self.context_var] = Calendar.objects.get_calendar_for_object(
            self.content_object.resolve(context), self.distinction
        )
        return ""


@register.tag
def get_calendar(parser, token):
    contents = token.split_contents()
    if len(contents) == 4:
        _, content_object, _, context_var = contents
        distinction = None
    elif len(contents) == 5:
        _, content_object, distinction, _, context_var = token.split_contents()
    else:
        raise template.TemplateSyntaxError(
            "%r tag follows form %r <content_object> as <context_var>"
            % (token.contents.split()[0], token.contents.split()[0])
        )
    return CalendarNode(content_object, distinction, context_var)


class CreateCalendarNode(template.Node):
    def __init__(self, content_object, distinction, context_var, name):
        self.content_object = template.Variable(content_object)
        self.distinction = distinction
        self.context_var = context_var
        self.name = name

    def render(self, context):
        context[self.context_var] = Calendar.objects.get_or_create_calendar_for_object(
            self.content_object.resolve(context), self.distinction, name=self.name
        )
        return ""


@register.tag
def get_or_create_calendar(parser, token):
    contents = token.split_contents()
    if len(contents) > 2:
        obj = contents[1]
        if "by" in contents:
            by_index = contents.index("by")
            distinction = contents[by_index + 1]
        else:
            distinction = None
        if "named" in contents:
            named_index = contents.index("named")
            name = contents[named_index + 1]
            if name[0] == name[-1]:
                name = name[1:-1]
        else:
            name = None
        if "as" in contents:
            as_index = contents.index("as")
            context_var = contents[as_index + 1]
        else:
            raise template.TemplateSyntaxError(
                "%r tag requires an a context variable: %r <content_object> [named <calendar name>] [by <distinction>] as <context_var>"
                % (token.split_contents()[0], token.split_contents()[0])
            )
    else:
        raise template.TemplateSyntaxError(
            "%r tag follows form %r <content_object> [named <calendar name>] [by <distinction>] as <context_var>"
            % (token.split_contents()[0], token.split_contents()[0])
        )
    return CreateCalendarNode(obj, distinction, context_var, name)


@register.simple_tag
def querystring_for_date(date, num=6):
    qs_parts = [
        ("year", date.year),
        ("month", date.month),
        ("day", date.day),
        ("hour", date.hour),
        ("minute", date.minute),
        ("second", date.second),
    ]
    query_string = "?" + urlencode(qs_parts[:num])
    # For compatibility with older Django versions, escape the
    # output. Starting with Django 1.9, simple_tags are automatically
    # passed through conditional_escape(). See:
    # https://docs.djangoproject.com/en/1.9/releases/1.9/#simple-tag-now-wraps-tag-output-in-conditional-escape
    return escape(query_string)


@register.simple_tag
def prev_url(target, calendar, period):
    now = timezone.now()
    delta = now - period.prev().start
    slug = calendar.slug
    if delta.total_seconds() > SCHEDULER_PREVNEXT_LIMIT_SECONDS:
        return ""
    context = {
        "url": "{}{}".format(
            reverse(target, kwargs={"calendar_slug": slug}),
            querystring_for_date(period.prev().start),
        )
    }
    return get_template("schedule/_prev.html").render(context)


@register.simple_tag
def next_url(target, calendar, period):
    now = timezone.now()
    slug = calendar.slug

    delta = period.next().start - now
    if delta.total_seconds() > SCHEDULER_PREVNEXT_LIMIT_SECONDS:
        return ""

    context = {
        "url": "{}{}".format(
            reverse(target, kwargs={"calendar_slug": slug}),
            querystring_for_date(period.next().start),
        )
    }
    return get_template("schedule/_next.html").render(context)


@register.inclusion_tag("schedule/_prevnext.html")
def prevnext(target, calendar, period, fmt=None):
    if fmt is None:
        fmt = settings.DATE_FORMAT
    context = {
        "calendar": calendar,
        "period": period,
        "period_name": format(period.start, fmt),
        "target": target,
    }
    return context


@register.inclusion_tag("schedule/_detail.html")
def detail(occurrence):
    context = {"occurrence": occurrence}
    return context


def _cook_slots(period, increment):
    """
    Prepare slots to be displayed on the left hand side
    calculate dimensions (in px) for each slot.
    Arguments:
    period - time period for the whole series
    increment - slot size in minutes
    """
    tdiff = datetime.timedelta(minutes=increment)
    num = int((period.end - period.start).total_seconds()) // int(tdiff.total_seconds())
    s = period.start
    slots = []
    for i in range(num):
        sl = period.get_time_slot(s, s + tdiff)
        slots.append(sl)
        s = s + tdiff
    return slots


@register.simple_tag
def hash_occurrence(occ):
    return "{}_{}".format(occ.start.strftime("%Y%m%d%H%M%S"), occ.event.id)



# Generated by Django 6.0 on 2026-01-06 01:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('schedule', '0014_use_autofields_for_pk'),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='calendarrelation',
            new_name='schedule_ca_content_cddadb_idx',
            old_fields=('content_type', 'object_id'),
        ),
        migrations.RenameIndex(
            model_name='event',
            new_name='schedule_ev_start_a258cb_idx',
            old_fields=('start', 'end'),
        ),
        migrations.RenameIndex(
            model_name='eventrelation',
            new_name='schedule_ev_content_6ceecb_idx',
            old_fields=('content_type', 'object_id'),
        ),
        migrations.RenameIndex(
            model_name='occurrence',
            new_name='schedule_oc_start_76a2f8_idx',
            old_fields=('start', 'end'),
        ),
    ]



import datetime

import pytz
from django.test import TestCase
from django.utils.html import escape

from schedule.models import Calendar, Event, Rule
from schedule.periods import Day, Period
from schedule.templatetags.scheduletags import (
    _cook_slots,
    create_event_url,
    next_url,
    prev_url,
    querystring_for_date,
)


class TestTemplateTags(TestCase):
    def setUp(self):
        self.day = Day(
            events=Event.objects.all(),
            date=datetime.datetime(
                datetime.datetime.now().year, 2, 7, 0, 0, tzinfo=pytz.utc
            ),
        )
        self.day_out_of_limit = Day(
            events=Event.objects.all(),
            date=datetime.datetime(
                datetime.datetime.now().year + 3, 2, 7, 0, 0, tzinfo=pytz.utc
            ),
        )
        self.day_out_of_limit_lower = Day(
            events=Event.objects.all(),
            date=datetime.datetime(
                datetime.datetime.now().year - 3, 2, 7, 0, 0, tzinfo=pytz.utc
            ),
        )

        rule = Rule.objects.create(frequency="WEEKLY")
        self.cal = Calendar.objects.create(name="MyCal", slug="MyCalSlug")

        Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(
                datetime.datetime.now().year, 1, 5, 8, 0, tzinfo=pytz.utc
            ),
            end=datetime.datetime(
                datetime.datetime.now().year, 1, 5, 9, 0, tzinfo=pytz.utc
            ),
            end_recurring_period=datetime.datetime(
                datetime.datetime.now().year, 5, 5, 0, 0, tzinfo=pytz.utc
            ),
            rule=rule,
            calendar=self.cal,
        )
        self.period = Period(
            events=Event.objects.all(),
            start=datetime.datetime(
                datetime.datetime.now().year, 1, 4, 7, 0, tzinfo=pytz.utc
            ),
            end=datetime.datetime(
                datetime.datetime.now().year, 1, 21, 7, 0, tzinfo=pytz.utc
            ),
        )

    def test_querystring_for_datetime(self):
        date = datetime.datetime(datetime.datetime.now().year, 1, 1, 0, 0, 0)
        query_string = querystring_for_date(date)
        self.assertEqual(
            escape(
                "?year={}&month=1&day=1&hour=0&minute=0&second=0".format(
                    datetime.datetime.now().year
                )
            ),
            query_string,
        )

    def test_prev_url(self):
        query_string = prev_url("month_calendar", self.cal, self.day)
        url_params = escape(
            "/calendar/month/MyCalSlug/?year={}&month=2&day=6&hour=0&minute=0&second=0".format(
                datetime.datetime.now().year
            )
        )
        expected = '<a href="{}"><span class="glyphicon glyphicon-circle-arrow-left"></span></a>'.format(
            url_params
        )
        self.assertEqual(query_string, expected)

    def test_next_url(self):
        query_string = next_url("month_calendar", self.cal, self.day)
        url_params = escape(
            "/calendar/month/MyCalSlug/?year={}&month=2&day=8&hour=0&minute=0&second=0".format(
                datetime.datetime.now().year
            )
        )
        expected = '<a href="{}"><span class="glyphicon glyphicon-circle-arrow-right"></span></a>'.format(
            url_params
        )
        self.assertEqual(query_string, expected)

    def test_next_url_upper_limit(self):
        query_string = next_url("month_calendar", self.cal, self.day_out_of_limit)
        self.assertEqual(query_string, "")

    def test_prev_url_lower_limit(self):
        query_string = prev_url("month_calendar", self.cal, self.day_out_of_limit_lower)
        self.assertEqual(query_string, "")

    def test_create_event_url(self):
        context = {}
        slot = self.period.get_time_slot(
            datetime.datetime(
                datetime.datetime.now().year, 1, 4, 7, 0, tzinfo=pytz.utc
            ),
            datetime.datetime(
                datetime.datetime.now().year, 1, 4, 7, 12, tzinfo=pytz.utc
            ),
        )
        query_string = create_event_url(context, self.cal, slot.start)
        expected = "/event/create/MyCalSlug/?year={}&month=1&day=4&hour=7&minute=0&second=0".format(
            datetime.datetime.now().year
        )
        self.assertEqual(query_string["create_event_url"], escape(expected))

    def test_all_day_event_cook_slots(self):
        start = datetime.datetime(
            datetime.datetime.now().year, 1, 5, 0, 0, tzinfo=pytz.utc
        )
        end = datetime.datetime(
            datetime.datetime.now().year, 1, 6, 0, 0, tzinfo=pytz.utc
        )
        event = Event.objects.create(
            title="All Day Event", start=start, end=end, calendar=self.cal
        )
        period = Day([event], start, end)

        slots = _cook_slots(period, 60)
        self.assertEqual(len(slots), 24)



import datetime

import pytz
from django.test import TestCase

from schedule.models import Calendar, Event, Rule
from schedule.models.events import Occurrence
from schedule.periods import Period


class TestOccurrence(TestCase):
    def setUp(self):
        rule = Rule.objects.create(frequency="WEEKLY")
        cal = Calendar.objects.create(name="MyCal")
        self.recurring_data = {
            "title": "Recent Event",
            "start": datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            "end": datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            "end_recurring_period": datetime.datetime(
                2008, 5, 5, 0, 0, tzinfo=pytz.utc
            ),
            "rule": rule,
            "calendar": cal,
        }
        self.recurring_event = Event.objects.create(**self.recurring_data)
        self.start = datetime.datetime(2008, 1, 12, 0, 0, tzinfo=pytz.utc)
        self.end = datetime.datetime(2008, 1, 27, 0, 0, tzinfo=pytz.utc)

    def test_persisted_occurrences(self):
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        with self.assertRaises(TypeError):
            hash(occurrences[0])
        persisted_occurrence = occurrences[0]
        persisted_occurrence.save()
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        self.assertTrue(occurrences[0].pk)
        self.assertFalse(occurrences[1].pk)

    def test_persisted_occurrences_are_hashable(self):
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        persisted_occurrence = occurrences[0]
        persisted_occurrence.save()
        self.assertEqual(hash(persisted_occurrence), persisted_occurrence.pk)

    def test_moved_occurrences(self):
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        moved_occurrence = occurrences[1]
        span_pre = (moved_occurrence.start, moved_occurrence.end)
        span_post = [x + datetime.timedelta(hours=2) for x in span_pre]
        # check has_occurrence on both periods
        period_pre = Period([self.recurring_event], span_pre[0], span_pre[1])
        period_post = Period([self.recurring_event], span_post[0], span_post[1])
        self.assertTrue(period_pre.has_occurrences())
        self.assertFalse(period_post.has_occurrences())
        # move occurrence
        moved_occurrence.move(
            moved_occurrence.start + datetime.timedelta(hours=2),
            moved_occurrence.end + datetime.timedelta(hours=2),
        )
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        self.assertTrue(occurrences[1].moved)
        # check has_occurrence on both periods (the result should be reversed)
        period_pre = Period([self.recurring_event], span_pre[0], span_pre[1])
        period_post = Period([self.recurring_event], span_post[0], span_post[1])
        self.assertFalse(period_pre.has_occurrences())
        self.assertTrue(period_post.has_occurrences())

    def test_cancelled_occurrences(self):
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        cancelled_occurrence = occurrences[2]
        cancelled_occurrence.cancel()
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        self.assertTrue(occurrences[2].cancelled)
        cancelled_occurrence.uncancel()
        occurrences = self.recurring_event.get_occurrences(
            start=self.start, end=self.end
        )
        self.assertFalse(occurrences[2].cancelled)

    def test_occurrence_eq_method(self):
        event2 = Event.objects.create(**self.recurring_data)
        self.assertEqual(
            self.recurring_event.get_occurrences(start=self.start, end=self.end)[0],
            event2.get_occurrences(start=self.start, end=self.end)[0],
        )
        self.assertNotEqual(
            self.recurring_event.get_occurrences(start=self.start, end=self.end)[0],
            event2.get_occurrences(start=self.start, end=self.end)[1],
        )
        self.assertNotEqual(
            self.recurring_event.get_occurrences(start=self.start, end=self.end)[0],
            event2,
        )

    def test_create_occurrence_without_event(self):
        """
        may be required for creating formsets, for example in admin
        """
        Occurrence()

    def test_get_occurrences_non_intersection_returns_empty_occ(self):
        rule = Rule.objects.create(frequency="DAILY")
        cal = Calendar.objects.create(name="MyCal", slug="mycal")
        recurring_event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2016, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2016, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2016, 8, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )
        occurrences = recurring_event.get_occurrences(start=self.start, end=self.end)
        self.assertEqual(occurrences, [])

    def test_get_occurrences_is_sorted(self):
        rule = Rule.objects.create(frequency="DAILY")
        cal = Calendar.objects.create(name="MyCal", slug="mycal")
        recurring_event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2016, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2016, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2016, 8, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )

        start = datetime.datetime(2016, 1, 12, 0, 0, tzinfo=pytz.utc)
        end = datetime.datetime(2016, 1, 27, 0, 0, tzinfo=pytz.utc)
        occurrences = recurring_event.get_occurrences(start=start, end=end)

        sorted_occurrences = sorted(occurrences, key=lambda occ: occ.start)

        self.assertEqual(occurrences, sorted_occurrences)



# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.request",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = "schedule.urls"

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "schedule",
    "tests",
)

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Internationalization

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

PASSWORD_HASHERS = ["django.contrib.auth.hashers.PBKDF2PasswordHasher"]

# URL prefix for static files.
STATIC_URL = "/static/"

STATICFILES_FINDERS = ("django.contrib.staticfiles.finders.AppDirectoriesFinder",)



import datetime

import pytz
from django.conf import settings
from django.db import connection
from django.test import TestCase
from django.test.utils import CaptureQueriesContext, override_settings

from schedule.models import Calendar, Event, Rule
from schedule.models.events import Occurrence
from schedule.periods import Day, Month, Period, Week, Year


class TestPeriod(TestCase):
    def setUp(self):
        rule = Rule.objects.create(frequency="WEEKLY")
        cal = Calendar.objects.create(name="MyCal")
        Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )
        self.period = Period(
            events=Event.objects.all(),
            start=datetime.datetime(2008, 1, 4, 7, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 21, 7, 0, tzinfo=pytz.utc),
        )

    def test_get_occurrences(self):
        occurrence_list = self.period.occurrences
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrence_list],
            [
                "2008-01-05 08:00:00+00:00 to 2008-01-05 09:00:00+00:00",
                "2008-01-12 08:00:00+00:00 to 2008-01-12 09:00:00+00:00",
                "2008-01-19 08:00:00+00:00 to 2008-01-19 09:00:00+00:00",
            ],
        )

    def test_nplus_one_queries_event_period(self):
        """Reproduces bug 420 occurences without title or desc will generate additional queries
        to retrieve the event model
        """
        rule = Rule.objects.create(frequency="WEEKLY")
        cal = Calendar.objects.get(name="MyCal")

        event = Event.objects.create(
            title="TEST",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2023, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )
        for _ in range(0, 21):
            # lets set occ without title and desc to use the event.title or event.desc
            Occurrence.objects.create(
                event=event,
                start=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
                end=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
                original_start=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
                original_end=datetime.datetime(2008, 1, 7, 8, 0, tzinfo=pytz.utc),
            )
        period = Period(
            events=[event],
            start=datetime.datetime(2008, 1, 4, 7, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2023, 1, 21, 7, 0, tzinfo=pytz.utc),
        )
        with CaptureQueriesContext(connection) as ctx:
            for occurrence in period.get_occurrences():
                pass

        executed_queries = len(ctx.captured_queries)
        assert executed_queries == 1, len(ctx.captured_queries)

    def test_get_occurrences_with_sorting_options(self):
        period = Period(
            events=Event.objects.all(),
            start=datetime.datetime(2008, 1, 4, 7, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 21, 7, 0, tzinfo=pytz.utc),
            sorting_options={"reverse": True},
        )
        occurrence_list = period.occurrences
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrence_list],
            [
                "2008-01-19 08:00:00+00:00 to 2008-01-19 09:00:00+00:00",
                "2008-01-12 08:00:00+00:00 to 2008-01-12 09:00:00+00:00",
                "2008-01-05 08:00:00+00:00 to 2008-01-05 09:00:00+00:00",
            ],
        )

    def test_get_occurrence_partials(self):
        occurrence_dicts = self.period.get_occurrence_partials()
        self.assertEqual(
            [
                (
                    occ_dict["class"],
                    occ_dict["occurrence"].start,
                    occ_dict["occurrence"].end,
                )
                for occ_dict in occurrence_dicts
            ],
            [
                (
                    1,
                    datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
                ),
                (
                    1,
                    datetime.datetime(2008, 1, 12, 8, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 12, 9, 0, tzinfo=pytz.utc),
                ),
                (
                    1,
                    datetime.datetime(2008, 1, 19, 8, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 19, 9, 0, tzinfo=pytz.utc),
                ),
            ],
        )

    def test_has_occurrence(self):
        self.assertTrue(self.period.has_occurrences())
        slot = self.period.get_time_slot(
            datetime.datetime(2008, 1, 4, 7, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 4, 7, 12, tzinfo=pytz.utc),
        )
        self.assertFalse(slot.has_occurrences())


class TestYear(TestCase):
    def setUp(self):
        self.year = Year(events=[], date=datetime.datetime(2008, 4, 1, tzinfo=pytz.utc))

    def test_get_months(self):
        months = self.year.get_months()
        self.assertEqual(
            [month.start for month in months],
            [datetime.datetime(2008, i, 1, tzinfo=pytz.utc) for i in range(1, 13)],
        )


class TestMonth(TestCase):
    def setUp(self):
        rule = Rule.objects.create(frequency="WEEKLY")
        cal = Calendar.objects.create(name="MyCal")
        Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )
        self.month = Month(
            events=Event.objects.all(),
            date=datetime.datetime(2008, 2, 7, 9, 0, tzinfo=pytz.utc),
        )

    def test_get_weeks(self):
        weeks = self.month.get_weeks()
        actuals = [(week.start, week.end) for week in weeks]

        if settings.FIRST_DAY_OF_WEEK == 0:
            expecteds = [
                (
                    datetime.datetime(2008, 1, 27, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 3, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 3, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 10, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 10, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 17, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 17, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 24, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 24, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 3, 2, 0, 0, tzinfo=pytz.utc),
                ),
            ]
        else:
            expecteds = [
                (
                    datetime.datetime(2008, 1, 28, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 4, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 4, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 11, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 11, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 18, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 18, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 25, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    datetime.datetime(2008, 2, 25, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 3, 3, 0, 0, tzinfo=pytz.utc),
                ),
            ]

        for actual, expected in zip(actuals, expecteds):
            self.assertEqual(actual, expected)

    def test_get_days(self):
        weeks = self.month.get_weeks()
        week = list(weeks)[0]
        days = week.get_days()
        actuals = [(len(day.occurrences), day.start, day.end) for day in days]

        if settings.FIRST_DAY_OF_WEEK == 0:
            expecteds = [
                (
                    0,
                    datetime.datetime(2008, 1, 27, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 28, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 28, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 29, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 29, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 30, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 30, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 31, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 31, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 1, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 2, 1, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 2, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    1,
                    datetime.datetime(2008, 2, 2, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 3, 0, 0, tzinfo=pytz.utc),
                ),
            ]

        else:
            expecteds = [
                (
                    0,
                    datetime.datetime(2008, 1, 28, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 29, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 29, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 30, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 30, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 1, 31, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 1, 31, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 1, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 2, 1, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 2, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    1,
                    datetime.datetime(2008, 2, 2, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 3, 0, 0, tzinfo=pytz.utc),
                ),
                (
                    0,
                    datetime.datetime(2008, 2, 3, 0, 0, tzinfo=pytz.utc),
                    datetime.datetime(2008, 2, 4, 0, 0, tzinfo=pytz.utc),
                ),
            ]

        for actual, expected in zip(actuals, expecteds):
            self.assertEqual(actual, expected)

    def test_month_convenience_functions(self):
        self.assertEqual(
            self.month.prev_month().start,
            datetime.datetime(2008, 1, 1, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            self.month.next_month().start,
            datetime.datetime(2008, 3, 1, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            self.month.current_year().start,
            datetime.datetime(2008, 1, 1, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            self.month.prev_year().start,
            datetime.datetime(2007, 1, 1, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            self.month.next_year().start,
            datetime.datetime(2009, 1, 1, 0, 0, tzinfo=pytz.utc),
        )


class TestDay(TestCase):
    def setUp(self):
        self.day = Day(
            events=Event.objects.all(),
            date=datetime.datetime(2008, 2, 7, 9, 0, tzinfo=pytz.utc),
        )

    def test_day_setup(self):
        self.assertEqual(
            self.day.start, datetime.datetime(2008, 2, 7, 0, 0, tzinfo=pytz.utc)
        )
        self.assertEqual(
            self.day.end, datetime.datetime(2008, 2, 8, 0, 0, tzinfo=pytz.utc)
        )

    def test_day_convenience_functions(self):
        self.assertEqual(
            self.day.prev_day().start,
            datetime.datetime(2008, 2, 6, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            self.day.next_day().start,
            datetime.datetime(2008, 2, 8, 0, 0, tzinfo=pytz.utc),
        )

    def test_time_slot(self):
        slot_start = datetime.datetime(2008, 2, 7, 13, 30, tzinfo=pytz.utc)
        slot_end = datetime.datetime(2008, 2, 7, 15, 0, tzinfo=pytz.utc)
        period = self.day.get_time_slot(slot_start, slot_end)
        self.assertEqual(period.start, slot_start)
        self.assertEqual(period.end, slot_end)

    def test_time_slot_with_dst(self):
        tzinfo = pytz.timezone("America/Vancouver")
        slot_start = datetime.datetime(2016, 3, 13, 0, 0, tzinfo=tzinfo)
        slot_end = datetime.datetime(2016, 3, 14, 0, 0, tzinfo=tzinfo)
        period = self.day.get_time_slot(slot_start, slot_end)
        self.assertEqual(period.start, slot_start)
        self.assertEqual(period.end, slot_end)

    def test_get_day_range(self):
        # This test exercises the case where a Day object is initiatized with
        # no date, which causes the Day constructor to call timezone.now(),
        # which always uses UTC.  This can cause a problem if the desired TZ
        # is not UTC, because the _get_day_range method typecasts the
        # tz-aware datetime to a naive datetime.

        # To simulate this case, we will create a NY tz date, localize that
        # date to UTC, then create a Day object with the UTC date and NY TZ

        NY = pytz.timezone("America/New_York")
        user_wall_time = datetime.datetime(2015, 11, 4, 21, 30, tzinfo=NY)
        timezone_now = user_wall_time.astimezone(pytz.utc)

        test_day = Day(events=Event.objects.all(), date=timezone_now, tzinfo=NY)

        expected_start = datetime.datetime(2015, 11, 4, 5, 00, tzinfo=pytz.utc)
        expected_end = datetime.datetime(2015, 11, 5, 5, 00, tzinfo=pytz.utc)

        self.assertEqual(test_day.start, expected_start)
        self.assertEqual(test_day.end, expected_end)


class TestOccurrencePool(TestCase):
    def setUp(self):
        rule = Rule.objects.create(frequency="WEEKLY")
        cal = Calendar.objects.create(name="MyCal")
        self.recurring_event = Event.objects.create(
            title="Recent Event",
            start=datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            end_recurring_period=datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule=rule,
            calendar=cal,
        )

    def testPeriodFromPool(self):
        """
        Test that period initiated with occurrence_pool returns the same occurrences as "straigh" period
        in a corner case whereby a period's start date is equal to the occurrence's end date
        """
        start = datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc)
        end = datetime.datetime(2008, 1, 5, 10, 0, tzinfo=pytz.utc)
        parent_period = Period(Event.objects.all(), start, end)
        period = Period(
            parent_period.events,
            start,
            end,
            parent_period.get_persisted_occurrences(),
            parent_period.occurrences,
        )
        self.assertEqual(parent_period.occurrences, period.occurrences)


class TestOccurrencesInTimezone(TestCase):
    def setUp(self):
        self.MVD = pytz.timezone("America/Montevideo")
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(
            frequency="DAILY", params="byweekday:SA", name="Saturdays"
        )
        Event.objects.create(
            title="Every Saturday Event",
            start=self.MVD.localize(datetime.datetime(2017, 1, 7, 22, 0)),
            end=self.MVD.localize(datetime.datetime(2017, 1, 7, 23, 0)),
            end_recurring_period=self.MVD.localize(datetime.datetime(2017, 2, 1)),
            rule=rule,
            calendar=cal,
        )

    @override_settings(TIME_ZONE="America/Montevideo")
    def test_occurrences_with_TZ(self):
        start = self.MVD.localize(datetime.datetime(2017, 1, 13))
        end = self.MVD.localize(datetime.datetime(2017, 1, 23))

        period = Period(Event.objects.all(), start, end, tzinfo=self.MVD)
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in period.occurrences],
            [
                "2017-01-14 22:00:00-03:00 to 2017-01-14 23:00:00-03:00",
                "2017-01-21 22:00:00-03:00 to 2017-01-21 23:00:00-03:00",
            ],
        )

    @override_settings(TIME_ZONE="America/Montevideo")
    def test_occurrences_sub_period_with_TZ(self):
        start = self.MVD.localize(datetime.datetime(2017, 1, 13))
        end = self.MVD.localize(datetime.datetime(2017, 1, 23))

        period = Period(Event.objects.all(), start, end, tzinfo=self.MVD)

        sub_start = self.MVD.localize(datetime.datetime(2017, 1, 13))
        sub_end = self.MVD.localize(datetime.datetime(2017, 1, 15))
        sub_period = period.get_time_slot(sub_start, sub_end)
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in sub_period.occurrences],
            ["2017-01-14 22:00:00-03:00 to 2017-01-14 23:00:00-03:00"],
        )


class TestWeeklyOccurrences(TestCase):
    def setUp(self):
        self.MVD = pytz.timezone("America/Montevideo")  # UTC-3
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY", name="daily")
        Event.objects.create(
            title="Test event",
            start=self.MVD.localize(datetime.datetime(2017, 1, 13, 15, 0)),
            end=self.MVD.localize(datetime.datetime(2017, 1, 14, 15, 0)),
            end_recurring_period=self.MVD.localize(datetime.datetime(2017, 1, 20)),
            rule=rule,
            calendar=cal,
        )

    def test_occurrences_inside_recurrence_period(self):
        start = self.MVD.localize(datetime.datetime(2017, 1, 13))

        period = Week(Event.objects.all(), start, tzinfo=self.MVD)

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in period.occurrences],
            [
                "2017-01-13 15:00:00-03:00 to 2017-01-14 15:00:00-03:00",
                "2017-01-14 15:00:00-03:00 to 2017-01-15 15:00:00-03:00",
            ],
        )

    def test_occurrences_outside_recurrence_period(self):
        start = self.MVD.localize(datetime.datetime(2017, 1, 23))

        period = Week(Event.objects.all(), start, tzinfo=self.MVD)

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in period.occurrences], []
        )

    def test_occurrences_no_end(self):
        event = Event.objects.filter(title="Test event").get()
        event.end_recurring_period = None
        start = self.MVD.localize(datetime.datetime(2018, 1, 13))

        period = Week([event], start, tzinfo=self.MVD)

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in period.occurrences],
            [
                "2018-01-06 15:00:00-03:00 to 2018-01-07 15:00:00-03:00",
                "2018-01-07 15:00:00-03:00 to 2018-01-08 15:00:00-03:00",
                "2018-01-08 15:00:00-03:00 to 2018-01-09 15:00:00-03:00",
                "2018-01-09 15:00:00-03:00 to 2018-01-10 15:00:00-03:00",
                "2018-01-10 15:00:00-03:00 to 2018-01-11 15:00:00-03:00",
                "2018-01-11 15:00:00-03:00 to 2018-01-12 15:00:00-03:00",
                "2018-01-12 15:00:00-03:00 to 2018-01-13 15:00:00-03:00",
                "2018-01-13 15:00:00-03:00 to 2018-01-14 15:00:00-03:00",
            ],
        )

    def test_occurrences_end_in_diff_tz(self):
        event = Event.objects.filter(title="Test event").get()
        AMSTERDAM = pytz.timezone("Europe/Amsterdam")
        # 2017-01-14 00:00 CET = 2017-01-13 21:00 UYT
        event.end_recurring_period = AMSTERDAM.localize(
            datetime.datetime(2017, 1, 14, 0, 0)
        )
        start = self.MVD.localize(datetime.datetime(2017, 1, 13))

        period = Week([event], start, tzinfo=self.MVD)

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in period.occurrences],
            ["2017-01-13 15:00:00-03:00 to 2017-01-14 15:00:00-03:00"],
        )


class TestAwareDay(TestCase):
    def setUp(self):
        self.timezone = pytz.timezone("Europe/Amsterdam")

        start = self.timezone.localize(datetime.datetime(2008, 2, 7, 0, 20))
        end = self.timezone.localize(datetime.datetime(2008, 2, 7, 0, 21))
        self.event = Event.objects.create(
            title="One minute long event on january seventh 2008 at 00:20 in Amsterdam.",
            start=start,
            end=end,
            calendar=Calendar.objects.create(name="MyCal"),
        )

        self.day = Day(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2008, 2, 7, 9, 0)),
            tzinfo=self.timezone,
        )

    def test_day_range(self):
        start = datetime.datetime(2008, 2, 6, 23, 0, tzinfo=pytz.utc)
        end = datetime.datetime(2008, 2, 7, 23, 0, tzinfo=pytz.utc)

        self.assertEqual(start, self.day.start)
        self.assertEqual(end, self.day.end)

    def test_occurrence(self):
        self.assertEqual(self.event in [o.event for o in self.day.occurrences], True)


class TestTzInfoPersistence(TestCase):
    def setUp(self):
        self.timezone = pytz.timezone("Europe/Amsterdam")
        self.day = Day(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

        self.week = Week(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

        self.month = Month(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

        self.year = Year(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

    def test_persistence(self):
        self.assertEqual(self.day.tzinfo, self.timezone)
        self.assertEqual(self.week.tzinfo, self.timezone)
        self.assertEqual(self.month.tzinfo, self.timezone)
        self.assertEqual(self.year.tzinfo, self.timezone)


class TestAwareWeek(TestCase):
    def setUp(self):
        self.timezone = pytz.timezone("Europe/Amsterdam")
        self.week = Week(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

    def test_week_range(self):
        start = self.timezone.localize(datetime.datetime(2013, 12, 15, 0, 0))
        end = self.timezone.localize(datetime.datetime(2013, 12, 22, 0, 0))

        self.assertEqual(self.week.tzinfo, self.timezone)
        self.assertEqual(start, self.week.start)
        self.assertEqual(end, self.week.end)


class TestAwareMonth(TestCase):
    def setUp(self):
        self.timezone = pytz.timezone("Europe/Amsterdam")
        self.month = Month(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 11, 17, 9, 0)),
            tzinfo=self.timezone,
        )

    def test_month_range(self):
        start = self.timezone.localize(datetime.datetime(2013, 11, 1, 0, 0))
        end = self.timezone.localize(datetime.datetime(2013, 12, 1, 0, 0))

        self.assertEqual(self.month.tzinfo, self.timezone)
        self.assertEqual(start, self.month.start)
        self.assertEqual(end, self.month.end)


class TestAwareYear(TestCase):
    def setUp(self):
        self.timezone = pytz.timezone("Europe/Amsterdam")
        self.year = Year(
            events=Event.objects.all(),
            date=self.timezone.localize(datetime.datetime(2013, 12, 17, 9, 0)),
            tzinfo=self.timezone,
        )

    def test_year_range(self):
        start = self.timezone.localize(datetime.datetime(2013, 1, 1, 0, 0))
        end = self.timezone.localize(datetime.datetime(2014, 1, 1, 0, 0))

        self.assertEqual(self.year.tzinfo, self.timezone)
        self.assertEqual(start, self.year.start)
        self.assertEqual(end, self.year.end)


class TestStrftimeRefactor(TestCase):
    """
    Test for the refactor of strftime
    """

    def test_years_before_1900(self):
        d = datetime.date(year=1899, month=1, day=1)
        m = Month([], d)
        try:
            m.name()
        except ValueError as value_error:
            self.fail(value_error)



import datetime

from django.test import TestCase
from django.utils import timezone

from schedule.models import Calendar, CalendarRelation, Event, Rule


class TestCalendarInheritance(TestCase):
    def test_get_or_create_calendar_for_object_when_proxy_calendar_should_return_proxy_calendar(
        self,
    ):
        class ProxyCalendar(Calendar):
            class Meta:
                proxy = True

        rule = Rule.objects.create()
        self.assertIsInstance(
            ProxyCalendar.objects.get_or_create_calendar_for_object(rule), ProxyCalendar
        )


class TestCalendar(TestCase):
    def test_get_recent_events_without_events_is_empty(self):
        calendar = Calendar()
        calendar.save()
        self.assertEqual(list(calendar.get_recent()), [])

    def test_get_recent_events_with_events_return_the_event(self):
        pass

    def test_occurrences_after_without_events_is_empty(self):
        calendar = Calendar()
        calendar.save()
        self.assertEqual(list(calendar.occurrences_after(timezone.now())), [])

    def test_occurrences_after_with_events_after_returns_events(self):
        calendar = Calendar.objects.create()
        start_after = timezone.now() + datetime.timedelta(days=1)
        end_after = start_after + datetime.timedelta(hours=1)
        event = Event.objects.create(
            title="Recent Event", start=start_after, end=end_after, calendar=calendar
        )
        calendar.events.add(event)
        occurrences = list(calendar.occurrences_after(timezone.now()))
        self.assertEqual(len(occurrences), 1)
        self.assertEqual(occurrences[0].start, start_after)
        self.assertEqual(occurrences[0].end, end_after)

    def test_occurrences_after_with_events_before_returns_empty(self):
        calendar = Calendar.objects.create()
        start_after = timezone.now() + datetime.timedelta(days=-1)
        end_after = start_after + datetime.timedelta(hours=1)
        event = Event.objects.create(
            title="Recent Event", start=start_after, end=end_after, calendar=calendar
        )
        calendar.events.add(event)
        occurrences = list(calendar.occurrences_after(timezone.now()))
        self.assertEqual(occurrences, [])

    def test_get_calendar_for_object(self):
        calendar = Calendar.objects.create(name="My Cal")
        rule = Rule.objects.create()
        calendar.create_relation(rule)
        result = Calendar.objects.get_calendar_for_object(rule)
        self.assertEqual(result.name, "My Cal")

    def test_get_calendar_for_object_without_calendars(self):
        with self.assertRaises(Calendar.DoesNotExist):
            rule = Rule.objects.create()
            Calendar.objects.get_calendar_for_object(rule)

    def test_get_calendar_for_object_with_more_than_one_calendar(self):
        calendar_1 = Calendar.objects.create(name="My Cal 1", slug="my-cal-1")
        calendar_2 = Calendar.objects.create(name="My Cal 2", slug="my-cal-2")
        rule = Rule.objects.create()
        calendar_1.create_relation(rule)
        calendar_2.create_relation(rule)
        with self.assertRaises(AssertionError):
            Calendar.objects.get_calendar_for_object(rule)

    def test_get_or_create_calendar_for_object_without_calendar(self):
        """
        Creation test
        """
        rule = Rule.objects.create()
        calendar = Calendar.objects.get_or_create_calendar_for_object(
            rule, name="My Cal"
        )
        self.assertEqual(calendar.name, "My Cal")
        calendar_from_rule = Calendar.objects.get_calendars_for_object(rule)[0]
        self.assertEqual(calendar, calendar_from_rule)

    def test_get_or_create_calendar_for_object_withouth_name(self):
        """
        Test with already created calendar
        """
        rule = Rule.objects.create()
        calendar = Calendar.objects.get_or_create_calendar_for_object(rule)
        calendar_from_rule = Calendar.objects.get_calendars_for_object(rule)[0]
        self.assertEqual(calendar, calendar_from_rule)

    def test_get_calendars_for_object_without_calendars(self):
        rule = Rule.objects.create()
        Calendar.objects.get_or_create_calendar_for_object(
            rule, name="My Cal", distinction="owner"
        )
        rule = Rule.objects.create()
        calendars = list(
            Calendar.objects.get_calendars_for_object(rule, distinction="owner")
        )
        self.assertEqual(len(calendars), 0)

    def test_calendar_absolute_and_event_url(self):
        """
        this test seems to not make too much send, just added since an
        url was with wrong reverse name.

        """
        rule = Rule.objects.create()
        calendar = Calendar.objects.get_or_create_calendar_for_object(
            rule, name="My Cal", distinction="owner"
        )
        calendar.get_absolute_url()
        CalendarRelation.objects.create_relation(calendar, rule)



import datetime

from django.test import TestCase

from schedule.forms import EventForm


class TestScheduleForms(TestCase):
    def test_event_form(self):
        now = datetime.datetime.now()
        data = {
            "start_0": now.strftime("%Y-%m-%d"),
            "start_1": "00:00",
            "end_0": (now + datetime.timedelta(days=1)).strftime("%Y-%m-%d"),
            "end_1": "00:00",
            "title": "some title",
        }
        form = EventForm(data=data)
        validated = form.is_valid()
        self.assertTrue(validated)

        data["end_0"] = (now - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
        form = EventForm(data=data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.non_field_errors()), 1)
        self.assertEqual(
            form.non_field_errors()[0], "The end time must be later than start time."
        )

        del data["end_0"]
        del data["end_1"]
        del data["start_0"]
        del data["start_1"]
        form = EventForm(data=data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.non_field_errors()), 0)
        self.assertEqual(len(form.errors), 2)
        self.assertEqual(len(form.errors["end"]), 1)
        self.assertEqual(form.errors["end"][0], "This field is required.")
        self.assertEqual(len(form.errors["start"]), 1)
        self.assertEqual(form.errors["start"][0], "This field is required.")



import datetime

import pytz
from django.contrib.auth.models import User
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse
from django.utils import timezone

from schedule.models import Calendar, Event, EventRelation, Rule


class TestEvent(TestCase):
    def __create_event(self, title, start, end, cal):
        return Event.objects.create(title=title, start=start, end=end, calendar=cal)

    def __create_recurring_event(self, title, start, end, end_recurring, rule, cal):
        return Event.objects.create(
            title=title,
            start=start,
            end=end,
            end_recurring_period=end_recurring,
            rule=rule,
            calendar=cal,
        )

    def test_edge_case_events(self):
        cal = Calendar.objects.create(name="MyCal")
        event_one = Event.objects.create(
            title="Edge case event test one",
            start=datetime.datetime(2013, 1, 5, 8, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            calendar=cal,
        )
        event_two = Event.objects.create(
            title="Edge case event test two",
            start=datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
            calendar=cal,
        )
        occurrences_two = event_two.get_occurrences(
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(1, len(occurrences_two))

        occurrences_one = event_one.get_occurrences(
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(0, len(occurrences_one))

    def test_recurring_event_get_occurrences(self):

        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        occurrences = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 12, 0, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 20, 0, 0, tzinfo=pytz.utc),
        )
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrences],
            [
                "2008-01-12 08:00:00+00:00 to 2008-01-12 09:00:00+00:00",
                "2008-01-19 08:00:00+00:00 to 2008-01-19 09:00:00+00:00",
            ],
        )

    def test_event_get_occurrences_after(self):

        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        event_one = self.__create_event(
            "Edge case event test one",
            datetime.datetime(2013, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            cal,
        )
        event_two = self.__create_event(
            "Edge case event test two",
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
            cal,
        )
        occurrences_two = event_two.get_occurrences(
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
        )

        self.assertEqual(1, len(occurrences_two))

        occurrences_one = event_one.get_occurrences(
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 12, 0, tzinfo=pytz.utc),
        )

        self.assertEqual(0, len(occurrences_one))

    def test_recurring_event_get_occurrences_2(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurring event test",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        occurrences = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 12, 0, 0, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 20, 0, 0, tzinfo=pytz.utc),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrences],
            [
                "2008-01-12 08:00:00+00:00 to 2008-01-12 09:00:00+00:00",
                "2008-01-19 08:00:00+00:00 to 2008-01-19 09:00:00+00:00",
            ],
        )

    def test_recurring_event_get_occurrences_after(self):

        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")
        recurring_event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        occurrences = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 5, tzinfo=pytz.utc),
            end=datetime.datetime(2008, 1, 6, tzinfo=pytz.utc),
        )
        occurrence = occurrences[0]
        occurrence2 = next(
            recurring_event.occurrences_after(
                datetime.datetime(2008, 1, 5, tzinfo=pytz.utc)
            )
        )
        self.assertEqual(occurrence, occurrence2)

    def test_recurring_event_with_moved_get_occurrences_after(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")
        recurring_event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        occurrence = recurring_event.get_occurrence(
            datetime.datetime(2008, 1, 12, 8, 0, tzinfo=pytz.utc)
        )
        occurrence.move(
            datetime.datetime(2008, 1, 15, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 15, 9, 0, tzinfo=pytz.utc),
        )
        gen = recurring_event.occurrences_after(
            datetime.datetime(2008, 1, 14, 8, 0, tzinfo=pytz.utc)
        )
        occurrence2 = next(gen)
        self.assertEqual(occurrence, occurrence2)

    def test_recurring_event_get_occurrence(self):

        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        occurrence = event.get_occurrence(
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc)
        )
        self.assertEqual(
            occurrence.start, datetime.datetime(2008, 1, 5, 8, tzinfo=pytz.utc)
        )
        occurrence.save()
        occurrence = event.get_occurrence(
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc)
        )
        self.assertTrue(occurrence.pk is not None)

    def test_prevent_type_error_when_comparing_naive_and_aware_dates(self):
        # this only test if the TypeError is raised
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 1, 5, 9, 0, tzinfo=pytz.utc),
            datetime.datetime(2008, 5, 5, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        naive_date = datetime.datetime(2008, 1, 20, 0, 0)
        self.assertIsNone(event.get_occurrence(naive_date))

    @override_settings(USE_TZ=False)
    def test_prevent_type_error_when_comparing_dates_when_tz_off(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2008, 1, 5, 8, 0),
            datetime.datetime(2008, 1, 5, 9, 0),
            datetime.datetime(2008, 5, 5, 0, 0),
            rule,
            cal,
        )
        naive_date = datetime.datetime(2008, 1, 20, 0, 0)
        self.assertIsNone(event.get_occurrence(naive_date))

    @override_settings(USE_TZ=False)
    def test_get_occurrences_when_tz_off(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurring event test",
            datetime.datetime(2008, 1, 5, 8, 0),
            datetime.datetime(2008, 1, 5, 9, 0),
            datetime.datetime(2008, 5, 5, 0, 0),
            rule,
            cal,
        )
        occurrences = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 12, 0, 0),
            end=datetime.datetime(2008, 1, 20, 0, 0),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrences],
            [
                "2008-01-12 08:00:00 to 2008-01-12 09:00:00",
                "2008-01-19 08:00:00 to 2008-01-19 09:00:00",
            ],
        )

    def test_event_get_ocurrence(self):
        cal = Calendar.objects.create(name="MyCal")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            cal,
        )
        occurrence = event.get_occurrence(start)
        self.assertEqual(occurrence.start, start)

    def test_occurrences_after_with_no_params(self):
        cal = Calendar.objects.create(name="MyCal")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            cal,
        )
        occurrences = list(event.occurrences_after())
        self.assertEqual(len(occurrences), 1)
        self.assertEqual(occurrences[0].start, start)
        self.assertEqual(occurrences[0].end, start + datetime.timedelta(hours=1))

    def test_occurrences_with_recurrent_event_end_recurring_period_edge_case(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_recurring_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(days=10),
            rule,
            cal,
        )
        occurrences = list(event.occurrences_after())
        self.assertEqual(len(occurrences), 11)

    def test_occurrences_with_recurrent_event_end_recurring_period_edge_case_max_loop_lower(
        self,
    ):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_recurring_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(days=10),
            rule,
            cal,
        )
        occurrences = list(event.occurrences_after(max_occurrences=4))
        self.assertEqual(len(occurrences), 4)

    def test_occurrences_with_recurrent_event_end_recurring_period_edge_case_max_loop_greater(
        self,
    ):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_recurring_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(days=10),
            rule,
            cal,
        )
        occurrences = list(event.occurrences_after(max_occurrences=20))
        self.assertEqual(len(occurrences), 11)

    def test_occurrences_with_recurrent_event_no_end_recurring_period_max_loop(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_recurring_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(hours=10),
            rule,
            cal,
        )
        occurrences = list(event.occurrences_after(max_occurrences=1))
        self.assertEqual(len(occurrences), 1)

    def test_get_for_object(self):
        user = User.objects.create_user("john", "lennon@thebeatles.com", "johnpassword")
        event_relations = list(
            Event.objects.get_for_object(user, "owner", inherit=False)
        )
        self.assertEqual(len(event_relations), 0)

        Rule.objects.create(frequency="DAILY")
        cal = Calendar.objects.create(name="MyCal")
        event = self.__create_event(
            "event test",
            datetime.datetime(2013, 1, 5, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2013, 1, 5, 9, 0, tzinfo=pytz.utc),
            cal,
        )
        events = list(Event.objects.get_for_object(user, "owner", inherit=False))
        self.assertEqual(len(events), 0)
        EventRelation.objects.create_relation(event, user, "owner")

        events = list(Event.objects.get_for_object(user, "owner", inherit=False))
        self.assertEqual(len(events), 1)
        self.assertEqual(event, events[0])

    def test_get_absolute(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="DAILY")
        start = timezone.now() + datetime.timedelta(days=1)
        event = self.__create_recurring_event(
            "Non recurring event test get_occurrence",
            start,
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(days=10),
            rule,
            cal,
        )
        url = event.get_absolute_url()
        self.assertEqual(reverse("event", kwargs={"event_id": event.id}), url)

    @override_settings(TIME_ZONE="Europe/Helsinki")
    def test_recurring_event_get_occurrence_in_timezone(self):
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        # Event start and end are UTC because that is what is coming
        # from the database
        event = self.__create_recurring_event(
            "Recurrent event test get_occurrence",
            datetime.datetime(2014, 3, 21, 6, 0, tzinfo=pytz.utc),
            datetime.datetime(2014, 3, 21, 8, 0, tzinfo=pytz.utc),
            datetime.datetime(2014, 4, 11, 0, 0, tzinfo=pytz.utc),
            rule,
            cal,
        )
        tzinfo = pytz.timezone("Europe/Helsinki")
        start = tzinfo.localize(datetime.datetime(2014, 3, 28, 8, 0))  # +2
        occurrence = event.get_occurrence(start)
        self.assertEqual(occurrence.start, start)
        occurrence.save()
        # DST change on March 30th from +2 to +3
        start = tzinfo.localize(datetime.datetime(2014, 4, 4, 8, 0))  # +3
        occurrence = event.get_occurrence(start)
        self.assertEqual(occurrence.start, start)

    def test_recurring_event_get_occurrence_different_end_timezone(self):
        end_recurring = datetime.datetime(2016, 7, 30, 11, 0, tzinfo=pytz.utc)

        event = self.__create_recurring_event(
            "Recurring event with end_reccurring_date in different TZ",
            datetime.datetime(2016, 7, 25, 10, 0, tzinfo=pytz.utc),
            datetime.datetime(2016, 7, 25, 11, 0, tzinfo=pytz.utc),
            end_recurring,
            Rule.objects.create(frequency="DAILY"),
            Calendar.objects.create(name="MyCal"),
        )
        tzinfo = pytz.timezone("Europe/Athens")
        occurrences = event.get_occurrences(
            tzinfo.localize(datetime.datetime(2016, 1, 1, 0, 0)),
            tzinfo.localize(datetime.datetime(2016, 12, 31, 23, 59)),
        )
        self.assertEqual(occurrences[-1].end, end_recurring)

    def test_recurring_event_get_occurrence_across_dst(self):

        pacific = pytz.timezone("US/Pacific")
        e_start = pacific.localize(datetime.datetime(2015, 3, 4, 9, 0))
        e_end = e_start
        recc_end = pacific.localize(datetime.datetime(2015, 3, 13, 9, 0))
        event = self.__create_recurring_event(
            "Recurring event with end_recurring_date that crosses a DST",
            e_start,
            e_end,
            recc_end,
            Rule.objects.create(frequency="WEEKLY"),
            Calendar.objects.create(name="MyCal"),
        )
        occs = event.get_occurrences(
            e_start, pacific.localize(datetime.datetime(2015, 3, 11, 10, 0))
        )
        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occs],
            [
                "2015-03-04 09:00:00-08:00 to 2015-03-04 09:00:00-08:00",
                "2015-03-11 09:00:00-07:00 to 2015-03-11 09:00:00-07:00",
            ],
        )

    @override_settings(USE_TZ=False)
    def test_get_occurrences_timespan_inside_occurrence(self):
        """
        Test whether occurrences are correctly obtained if selected timespan start
        and end happen completely inside an occurrence.
        """
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurring event test",
            datetime.datetime(2008, 1, 5, 8, 0),
            datetime.datetime(2008, 1, 5, 9, 0),
            datetime.datetime(2008, 5, 5, 0, 0),
            rule,
            cal,
        )
        occurrences = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 12, 8, 15),
            end=datetime.datetime(2008, 1, 12, 8, 30),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occurrences],
            ["2008-01-12 08:00:00 to 2008-01-12 09:00:00"],
        )

    @override_settings(USE_TZ=False)
    def test_get_occurrences_timespan_partially_inside_occurrence(self):
        """
        Test whether occurrences are correctly obtained if selected timespan start
        outside of timepan but ends inside occurrence.
        """
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurring event test",
            datetime.datetime(2008, 1, 5, 8, 0),
            datetime.datetime(2008, 1, 5, 9, 0),
            datetime.datetime(2008, 5, 5, 0, 0),
            rule,
            cal,
        )

        occs1 = recurring_event.get_occurrences(
            start=datetime.datetime(2006, 1, 1, 0, 0),
            end=datetime.datetime(2008, 1, 19, 8, 30),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occs1],
            [
                "2008-01-05 08:00:00 to 2008-01-05 09:00:00",
                "2008-01-12 08:00:00 to 2008-01-12 09:00:00",
                "2008-01-19 08:00:00 to 2008-01-19 09:00:00",
            ],
        )

        occs2 = recurring_event.get_occurrences(
            start=datetime.datetime(2006, 1, 1, 0, 0),
            end=datetime.datetime(2008, 1, 19, 10, 0),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occs2],
            [
                "2008-01-05 08:00:00 to 2008-01-05 09:00:00",
                "2008-01-12 08:00:00 to 2008-01-12 09:00:00",
                "2008-01-19 08:00:00 to 2008-01-19 09:00:00",
            ],
        )

    @override_settings(USE_TZ=False)
    def test_get_occurrences_timespan_edge_cases(self):
        """
        Test whether occurrence list get method behave when requesting them on
        timespan limits
        """
        cal = Calendar.objects.create(name="MyCal")
        rule = Rule.objects.create(frequency="WEEKLY")

        recurring_event = self.__create_recurring_event(
            "Recurring event test",
            datetime.datetime(2008, 1, 5, 8, 0),
            datetime.datetime(2008, 1, 5, 9, 0),
            datetime.datetime(2008, 5, 5, 0, 0),
            rule,
            cal,
        )

        occs1 = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 5, 8, 0),
            end=datetime.datetime(2008, 1, 5, 8, 30),
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occs1],
            ["2008-01-05 08:00:00 to 2008-01-05 09:00:00"],
        )

        occs2 = recurring_event.get_occurrences(
            start=datetime.datetime(2008, 1, 5, 7, 0),
            end=datetime.datetime(2008, 1, 5, 8, 0),
        )

        self.assertEqual(["{} to {}".format(o.start, o.end) for o in occs2], [])

    @override_settings(USE_TZ=False)
    def test_get_occurrences_rule_weekday_param(self):
        """
        Test whether occurrence list get method behaves correctly while using
        complex rule parameters
        """
        cal = Calendar.objects.create(name="MyCal")
        # Last Fridays of each month
        rule = Rule.objects.create(
            frequency="MONTHLY", params="BYWEEKDAY:FR;BYSETPOS:-1"
        )

        recurring_event = self.__create_recurring_event(
            "Last Friday of each month",
            datetime.datetime(2016, 11, 1),
            datetime.datetime(2016, 11, 1),
            datetime.datetime(2030, 1, 1),
            rule,
            cal,
        )

        occs = recurring_event.get_occurrences(
            start=datetime.datetime(2016, 11, 1), end=datetime.datetime(2017, 12, 31)
        )

        self.assertEqual(
            ["{} to {}".format(o.start, o.end) for o in occs],
            [
                "2016-11-25 00:00:00 to 2016-11-25 00:00:00",
                "2016-12-30 00:00:00 to 2016-12-30 00:00:00",
                "2017-01-27 00:00:00 to 2017-01-27 00:00:00",
                "2017-02-24 00:00:00 to 2017-02-24 00:00:00",
                "2017-03-31 00:00:00 to 2017-03-31 00:00:00",
                "2017-04-28 00:00:00 to 2017-04-28 00:00:00",
                "2017-05-26 00:00:00 to 2017-05-26 00:00:00",
                "2017-06-30 00:00:00 to 2017-06-30 00:00:00",
                "2017-07-28 00:00:00 to 2017-07-28 00:00:00",
                "2017-08-25 00:00:00 to 2017-08-25 00:00:00",
                "2017-09-29 00:00:00 to 2017-09-29 00:00:00",
                "2017-10-27 00:00:00 to 2017-10-27 00:00:00",
                "2017-11-24 00:00:00 to 2017-11-24 00:00:00",
                "2017-12-29 00:00:00 to 2017-12-29 00:00:00",
            ],
        )


class TestEventRelationManager(TestCase):
    def test_get_events_for_object(self):
        pass
